/**
 * @file qrsdetector.hpp
 * @author Lars Johannesen <lj@halloffame.dk>
 *
 * @section LICENSE
 * This program is licensed under LGPL3, see LICENSE
 * 
 * @section DESCRIPTION
 * The classes in this file are used for describing characteristics about ecg records
 */
#ifndef ECGLIB_QRSDETECTOR_LJ_2014_06_06
#define ECGLIB_QRSDETECTOR_LJ_2014_06_06 1

#include <ecglib_bits/qrsdetectors/detector.hpp>
#include <ecglib_bits/qrsdetectors/detail/detail.hpp>
#include <ecglib_bits/util/config.hpp>

#include <boost/circular_buffer.hpp>

namespace ecglib {
	class qrsdetector_config : public config {
		public:
			qrsdetector_config() : config() {
				defaults();
			}

		protected:
			void defaults();
	};

	/**
	* @brief QRS peak and location
	*/
	struct qrsdetector_peak {
		std::size_t loc;
		double value;

		/**
		* @brief Empty peak
		*/
		qrsdetector_peak() {
		}

		/**
		* @brief Set peak/loc
		*
		* @param l Peak location
		* @param val Peak value
		*/
		qrsdetector_peak(const std::size_t l, const double val) : loc(l), value(val) {
		}
	};

	/**
	* @brief QRSdetector state
	*/
	class qrsdetector_state {
		public:
			/**
			* @brief Initialize using the config
			*
			* @param cfg Config
			*/
			qrsdetector_state(const qrsdetector_config &cfg);

		public:
			/**
			* @brief Detect a signal peak at location loc with value val
			*
			* @param loc Signal peak location
			* @param val Signal value 
			*/
			void detect_speak(const std::size_t loc, const double val);

			/**
			* @brief Detect a noise peak at a location loc with value val
			*
			* @param loc Noise peak location
			* @param val Noise value 
			*/
			void detect_npeak(const std::size_t loc, const double val);

			/**
			* @brief Get noise npeak num
			*
			* @param num Noise peak number
			*
			* @return Noise peak
			*/
			qrsdetector_peak npeak(const std::size_t num) const {
				// FIXME: Check?
				return _npeaks[num];
			}

			/**
			* @brief Get signal peak num
			*
			* @param num Signal peak number
			*
			* @return Signal peak
			*/
			qrsdetector_peak speak(const std::size_t num) const {
				// FIXME: Check?
				return _speaks[num];
			}

			/**
			* @brief Number of signal peaks
			*
			* @return Number of signal peaks
			*/
			std::size_t nspeak() const {
				return _speaks.size();
			}

			/**
			* @brief Number of noise peaks
			*
			* @return Number of noise peaks
			*/
			std::size_t nnpeak() const {
				return _npeaks.size();
			}

			/**
			* @brief Get number of maximum noise peak in some interval ( [begin, end] )
			*
			* @param begin Beginning of search interval
			* @param end End of search interval
			*
			* @return Number of maximum noise peak
			*/
			std::size_t max_npeak(const std::size_t begin, const std::size_t end) const;

			/**
			* @brief Reclassify a noise peak as a signal peak
			*
			* @param num Noise peak number
			*/
			void reclassify_npeak(const std::size_t num) {
				// FIXME check

				qrsdetector_peak p = npeak(num);

				boost::circular_buffer<qrsdetector_peak>::iterator it = _npeaks.begin() + num;
				_npeaks.erase(it);

				detect_speak(p.loc, p.value);
			}

			/**
			* @brief Increases number of missed detections if we have not been saturated
			*/
			void missed() {
				_missed = _missed == _maxmissed ? _missed : _missed + 1;
				_thrpenalty = 1.-((_maxthrlower * _missed) / _maxmissed);
			}
	
			/**
			* @brief Sets the number of missed detections 
			*
			* @param missed New number of missed detections
			*/
			void missed(const std::size_t missed) {
				// FIXME check
				_missed = 0;
				_thrpenalty = 1;
			}

			/**
			* @brief Get threshold penalty fraction
			*
			* @return Threshold penalty fraction
			*/
			double penalty() const {
				return _thrpenalty;
			}

			/**
			* @brief Get number of missed detections
			*
			* @return Number of missed detections
			*/
			std::size_t nmissed() const {
				return _missed;
			}

			/**
			* @brief Force update of signal level
			*
			* @param val Signal peak level
			*/
			void spki(const double val) {
				_spki = val;

				update_thresholds();
			}

			/**
			* @brief Get value of signal level
			*
			* @return Signal level
			*/
			double spki() const {
				return _spki;
			}

			/**
			* @brief Force update of noise level
			*
			* @param val Noise level
			*/
			void npki(const double val) {
				_npki = val;

				update_thresholds();
			}

			/**
			* @brief Get noise level
			*
			* @return Noise level
			*/
			double npki() const {
				return _npki;
			}

			/**
			* @brief Flag a candidate peak
			*
			* @param loc Location
			* @param left Left 
			* @param right Right
			*/
			void location(const std::size_t loc, const std::size_t left, const std::size_t right) {
				_loc = loc;
				_left = left;
				_right = right;
				_ready = true;
			}

			/**
			* @brief Get left location of current candidate
			*
			* @return Left location
			*/
			std::size_t left() const {
				return _left;
			}

			/**
			* @brief Get right location of current candidate
			*
			* @return Right location
			*/
			std::size_t right() const {
				return _right;
			}

			/**
			* @brief Get location of current candidate
			*
			* @return Candidate location
			*/
			std::size_t location() const {
				return _loc;
			}

			/**
			* @brief Get if we are ready for detection
			*
			* @return True if ready
			*/
			bool isready() const {
				return _ready;
			}

			/**
			* @brief Set ready flag
			*
			* @param ready Ready value
			*/
			void ready(const bool ready) {
				_ready = ready;
			}

			/**
			* @brief Get primary threshold (adjusted for threshold penalty, i.e. lowered if we have missing beats)
			*
			* @return Primary threshold
			*/
			double thr1() const {
				return _thr1*_thrpenalty;
			}

			/**
			* @brief Get secondary threshold (adjusted for threshold penalty, i.e. lowered if we have missing beats)
			*
			* @return Secondary threshold
			*/
			double thr2() const {
				return _thr2*_thrpenalty;
			}

			/**
			* @brief Get ratio between estimated signal and noise level
			*
			* @return Signal to noise ratio
			*/
			double snr() const {
				if(_npki <= 0) {
					return 1;
				}

				return _spki/_npki;
			}

		private:
			/**
			* @brief Update primary and secondary detection thresholds
			*/
			void update_thresholds() {
				_thr1 = _npki + _thr1frac * (_spki - _npki);
				_thr2 = _thr1 * _thr2frac;
			}	

		private:
			/**
			* @brief Signal level
			*/
			double _spki;

			/**
			* @brief Noise level
			*/
			double _npki;

			/**
			* @brief Circular buffer of signal peaks
			*/
			boost::circular_buffer<qrsdetector_peak> _speaks;

			/**
			* @brief Circular buffer of noise peaks
			*/
			boost::circular_buffer<qrsdetector_peak> _npeaks;

			/**
			* @brief Fraction used to update signal/noise levels (weight of new peak)
			*/
			double _frac1;

			/**
			* @brief Fraction to update signal noise/levels (1-frac1), weight of old peaks
			*/
			double _frac2;

			/**
			* @brief Primary threshold
			*/
			double _thr1;

			/**
			* @brief Secondary threshold
			*/
			double _thr2;

			/**
			* @brief Fraction to float primary level above noise level
			*/
			double _thr1frac;

			/**
			* @brief Fraction of primary threshold, used to define secondary level
			*/
			double _thr2frac;
	
			/**
			* @brief If we are ready for a detection
			*/
			bool _ready;

			/**
			* @brief Current candidate location
			*/
			std::size_t _loc;

			/**
			* @brief Current candidate left location
			*/
			std::size_t _left;

			/**
			* @brief Current candidate right location
			*/
			std::size_t _right;

			/**
			* @brief Maximum relative change in signal level
			*/
			double _maxsigchg;

			/**
			* @brief Maximum relative change in noise level
			*/
			double _maxnoisechg;

			/**
			* @brief Max number of missed detections
			*/
			std::size_t _maxmissed;

			/**
			* @brief Current number of missed detections
			*/
			std::size_t _missed;

			/**
			* @brief Max lowering of threshold due to missed detection (fraction of original threshold prior to missed detections)
			*/
			double _maxthrlower;

			/**
			* @brief Current threshold penalty (based on missed detections)
			*/
			double _thrpenalty;
	};

	/**
	* @brief QRS training policy, more liberal than main detection policy but not as liberal as the default detection policy (grabs everything more or less)
	*/
	class qrstrainingpolicy {
		public:
			typedef ecglib::pointmap annotation;

		public:
			/**
			* @brief Setup training policy
			*
			* @param nchannels Number of channels
			* @param cfg Configuration
			*/
			qrstrainingpolicy(const std::size_t nchannels, const qrsdetector_config &cfg)  : _nch(nchannels) {
				_minleft = cfg.get<int>("minleft");
				_minright = cfg.get<int>("minright");
				_minwidth = cfg.get<int>("minwidth");
				_maxwidth = cfg.get<int>("maxwidth");
				_maxmove = cfg.get<int>("maxmove");
				_frac = cfg.get<double>("detfrac");

				_ready = false;

				_peakloc.resize(_nch);
				_peakready.resize(_nch);

				for(std::size_t i = 0; i < _nch; ++i) {
					_peakready[i] = false;
				}
			}

		public:
			/**
			* @brief Attempt to detect a candidate peak
			*
			* @tparam data Data type
			* @param state Current detectorstate
			* @param input Input data
			* @param ch Channel we are detecting in
			* @param idx Index of candidate
			* @param fs Sampling frequency
			*/
			template<class data>
			void cpeak(detectorstate &state, const data &input, const std::size_t ch, const std::size_t idx, const double fs) {
				try {
					const int maxmove = static_cast<int>(round(_maxmove * fs/1000.));
					const int maxwidth = static_cast<int>(round(_maxwidth * fs/1000.));
					const int minwidth = static_cast<int>(round(_minwidth * fs/1000.));
					const int minleft = static_cast<int>(round(_minleft * fs/1000.));
					const int minright = static_cast<int>(round(_minright * fs/1000.));

					auto res = findcog(input.col(ch), _frac, _sfrac, state.cpeak(ch));
					int loc = std::get<0>(res);
					int left = std::get<1>(res);
					int right = std::get<2>(res);

					double pval = input(loc, ch);

					int width = right - left;
					int leftwidth = loc - left;
					int rightwidth = right - loc;

					if( abs(loc-static_cast<int>(state.cpeak(ch))) < maxmove &&
						(width > minwidth && width < maxwidth) &&
						leftwidth > minleft &&
						rightwidth > minright
					  ) {
						_ready = true;

						if (!_peakready[ch]) {
							_peakready[ch] = true;
							_peakloc[ch] = loc;
						} else {
							// NOTE: Lets keep tallest
							if(input[_peakloc[loc]] < pval) {
								_peakloc[ch] = loc;
							}
						}

						state.cpeak(ch, idx);
					}
				} catch (...) {
					// Note: Nothing needs to happen if it is not possible to determine COG for the basic policy
					state.cpeak(ch, idx);
				}
			}

			/**
			* @brief Detect peak if ready
			*
			* @tparam data Data type
			* @param pm Annotation type
			* @param idx Current index in data
			* @param state Detector state
			* @param input Input data
			* @param fs Sampling frequency
			*/
			template<class data>
			void peak(annotation &pm, const std::size_t idx, detectorstate &state, const data &input, const double fs) {
				if(_ready) {
					for(std::size_t i = 0; i < _nch; ++i) {
						if(_peakready[i]) {
							pm[i][_peakloc[i]] = ecglib::annotation(_peakloc[i],ecglib::annotation_type::RPEAK,i);
							_peakready[i] = false;
						}
					}

					_ready = false;
				}
			}

		protected:
			/**
			* @brief Number of channels
			*/
			std::size_t _nch;

			/**
			* @brief Peak locations by channel
			*/
			std::vector<std::size_t> _peakloc;

			/**
			* @brief Peak detected by channel
			*/
			std::vector<bool> _peakready;

			/**
			* @brief Maximum distance we can move when determining center of gravity (ms)
			*/
			std::size_t _maxmove;

			/**
			* @brief Minimum width (ms)
			*/
			std::size_t _minwidth;

			/**
			* @brief Maximum width (ms)
			*/
			std::size_t _maxwidth;

			/**
			* @brief Minimum left width (ms)
			*/
			std::size_t _minleft;

			/**
			* @brief Minimum right width (ms)
			*/
			std::size_t _minright;

			/**
			* @brief U3 peak fraction
			*/
			double _frac;

			/**
			* @brief U3 slope fraction
			*/
			double _sfrac;

			/**
			* @brief Overall ready
			*/
			bool _ready;
	};

	/**
	* @brief Main QRS detector logic
	*
	* @tparam rrpredpolicy RR predictor to use
	*/
	template<class rrpredpolicy>
	class qrsdetectorpolicy {
		public:
			typedef ecglib::pointmap annotation;

		public:
			/**
			* @brief QRS detector logic
			*
			* @param rrpred RRprediction logic
			* @param nchannels Number of channels
			* @param cfg Configuration
			*/
			qrsdetectorpolicy(const rrpredpolicy rrpred, const std::size_t nchannels, const qrsdetector_config &cfg) : _nch(nchannels), _ready(false), _rrpred(rrpred), _state(nchannels, qrsdetector_state(cfg)) {
				_minrr = cfg.get<int>("minrr");
				_maxmove = cfg.get<int>("maxmove");
				_minval = cfg.get<double>("minval");

				_minleft = cfg.get<int>("minleft");
				_minright = cfg.get<int>("minright");
				_minwidth = cfg.get<int>("minwidth");
				_maxwidth = cfg.get<int>("maxwidth");

				_maxdet = cfg.get<double>("maxdet");
				_frac = cfg.get<double>("frac");
				_sfrac = cfg.get<double>("slopefrac");

				_searchw = cfg.get<int>("searchwindow");
				_minsnr = cfg.get<double>("minsnr");

				_degrad = cfg.get<double>("thrdegrad");

				_nreq = static_cast<int>(floor(cfg.get<double>("nreq") * nchannels));
				_maxinitmissed = cfg.get<int>("maxinitmissed");
				_missedtime = cfg.get<int>("missedtime");
				_detectonoff = cfg.get<int>("detectonoff");

				_firstdet = true;
				_lastloc = 0;
			}

		public:
			/**
			* @brief Determine if current candidate peak is good
			*
			* @tparam data Data type
			* @param state Detector state
			* @param input Input data
			* @param ch Current channel
			* @param idx Index of current candidate
			* @param fs Sampling frequency
			*
			* @return True if it was detected
			*/
			template<class data>
			bool cpeak(detectorstate &state, const data &input, const std::size_t ch, const std::size_t idx, const double fs) {
				bool ores = false;

				try {
					const int maxmove = static_cast<int>(round(_maxmove * fs/1000.));
					const int maxwidth = static_cast<int>(round(_maxwidth * fs/1000.));
					const int minwidth = static_cast<int>(round(_minwidth * fs/1000.));
					const int minleft = static_cast<int>(round(_minleft * fs/1000.));
					const int minright = static_cast<int>(round(_minright * fs/1000.));

					auto res = findcog(input.col(ch), _frac, _sfrac,state.cpeak(ch));
					int loc = std::get<0>(res);
					int left = std::get<1>(res);
					int right = std::get<2>(res);

					double pval = input(loc, ch);

					int width = right - left;
					int leftwidth = loc - left;
					int rightwidth = right - loc;

					if(pval > _state[ch].thr1() && 
						abs(loc-static_cast<int>(state.cpeak(ch))) < maxmove &&
						(width > minwidth && width < maxwidth) &&
						leftwidth > minleft &&
						rightwidth > minright &&
						((_state[ch].spki() == 0) || pval < (_state[ch].spki() * _maxdet)) &&
					  	 pval > _minval
					  )
					{
						if (!_state[ch].isready()) {
							_state[ch].detect_speak(loc, input(loc,ch));

							if(!_ready && (_nch==1 || (_state[ch].snr() > _minsnr && _state[ch].nmissed() < _maxinitmissed))) {
								_ready = true;
								_startloc = loc;
							}

							_state[ch].location(loc, left, right);
							_state[ch].missed(0);
							ores = true;
						} else {
							// FIXME: This needs to be improved, current strategy is to pick tallest
							if(input(_state[ch].location(), ch) > input(loc,ch)) {
								_state[ch].detect_npeak(loc, input(loc, ch));
							} else {
								_state[ch].detect_speak(loc, input(loc,ch));
								_state[ch].detect_npeak(loc, input(_state[ch].location(), ch));
								_state[ch].location(loc, left, right);
								ores = true;
							}
						}
						state.cpeak(ch, idx);
					} else {
						// NOTE: If the peak is not appropriate for a signal peak use it to update noise
						// but only if it is below threshold
						if(input(loc,ch) < _state[ch].thr1()) {
							_state[ch].detect_npeak(loc, input(loc, ch));
						}

						state.cpeak(ch, idx);
					}
				} catch (...) {
					//NOTE: Peaks for which left/right could not be located should just be ignored
					state.cpeak(ch, idx);
				}

				return ores;
			}

			/**
			* @brief Determine if we have a detected beat
			*
			* @tparam data Input data type
			* @param pm Pointmap
			* @param idx Current index
			* @param state Detector state
			* @param input Input data
			* @param fs Sampling frequency
			*/
			template<class data>
			void peak(annotation &pm, const std::size_t idx, detectorstate &state, const data &input, const double fs) {
				const std::size_t minrr_samp = static_cast<std::size_t>(_minrr * fs/1000.);
				const std::size_t missedtime_samp = static_cast<std::size_t>(_missedtime * fs/1000.);

				// Case1: We have waited for minrr since the detection was triggered for any lead
				if(_ready && (idx - _startloc) > minrr_samp) {
					arma::vec locs = arma::zeros<arma::vec>(_nch);
					unsigned int nready = 0;

					for(std::size_t i = 0; i < _nch; ++i) {
						if(_state[i].isready()) {
							locs(i) = _state[i].location();
							unsigned int tloc = static_cast<unsigned int>(round(locs(i) * fs / 1000.));
							pm[i][tloc] = ecglib::annotation(tloc,ecglib::annotation_type::RPEAK,i);
							// FIXME: Qon/Qoff
							nready++;
						}
					}

					uvec lidx = find(locs>0);
					vec tmp = detail::grab(locs, lidx);
					std::size_t gloc = median(tmp);

					// Did we miss other leads?
					if(nready < _nch) {
						const std::size_t searchw = static_cast<std::size_t>(_searchw * fs/1000.);
						for(std::size_t i = 0; i < _nch; ++i) {
							if(!_state[i].isready()) {
								int ccpeak = state.cpeak(i);
								bool found = false;

								// Strategy 1: Check current candidate peak
								if(abs(ccpeak-gloc) < searchw) {
									// If the current candidate peak is within search window
									// Attempt to detect it
									if(cpeak(state, input, i, ccpeak, fs)) {
										locs(i) = _state[i].location();
										unsigned int tloc = static_cast<unsigned int>(round(locs(i) * fs / 1000.));
										pm[i][tloc] = ecglib::annotation(tloc,ecglib::annotation_type::RPEAK,i);
										// FIXME: Qon/Qoff

										found = false;		
									}
								}

								// Strategy 2: Look to see if any noise peaks are acceptable (>thr2)	
								try {
									std::size_t num = _state[i].max_npeak(gloc-searchw, gloc+searchw);
									qrsdetector_peak maxnpeak = _state[i].npeak(num);
									
									if(maxnpeak.value > (_state[i].thr2() * _state[i].penalty()) && maxnpeak.value > _minval) {
										_state[i].reclassify_npeak(num);
						
										auto res = findcog(input.col(i), _frac, _sfrac, ccpeak);
										int loc = std::get<0>(res);

										locs(i) = loc;
										unsigned int tloc = static_cast<unsigned int>(round(locs(i) * fs / 1000.));
										pm[i][tloc] = ecglib::annotation(tloc,ecglib::annotation_type::RPEAK,i);
										// FIXME: Qon/Qoff
									}
								} catch (...) {
									// Note: Nothing was found
								}
							} // End if lead not ready
						} // End foreach lead
					} // End if we are missing leads
				
					lidx = find(locs>0);
					tmp = detail::grab(locs, lidx);
					gloc = median(tmp);
					_lastloc = gloc;
					_firstdet = false;

					unsigned int tloc = static_cast<unsigned int>(round(gloc * 1000./fs));
					pm[ecglib::GLOBAL_LEAD][tloc] = ecglib::annotation(tloc, ecglib::annotation_type::RPEAK, ecglib::GLOBAL_LEAD);
					// FIXME: Qon/Qoff

					for(std::size_t i = 0; i < _nch; ++i) {
						if(locs(i) == 0) {
							_state[i].missed();
						}
						_state[i].ready(false);
					}

					_ready = false;
					_rrpred.detectqrs(gloc, fs);
				// Case2: We have not detected a beat where we should have according to RR history
				} else if(!_ready && _rrpred.ready() && idx > _rrpred.nextqrs(fs)) {
					arma::vec locs = arma::zeros<arma::vec>(_nch);
					unsigned int nready = 0;
			
					auto predpos = _rrpred.predicted_location(fs);
					const std::size_t lbound = std::get<0>(predpos);
					const std::size_t ubound = std::get<1>(predpos);

					for(std::size_t i = 0; i < _nch; ++i) {
						int cpeak = state.cpeak(i);
						int num = -1;
						qrsdetector_peak maxnpeak;

						try {
							num = _state[i].max_npeak(lbound, ubound);
							maxnpeak = _state[i].npeak(num);
						} catch (...) {

						}

						// Strategy 1: Check current candidate peak
						if(cpeak > lbound && cpeak < ubound &&
						 	 input(cpeak, i) > _state[i].thr2() &&
							 (num == -1 || input(cpeak,i) > maxnpeak.value) && input(cpeak,i) > _minval
							 ) {

							_state[i].detect_speak(cpeak, input(cpeak, i));
							auto res = findcog(input.col(i), _frac, _sfrac, cpeak);

							locs(i) = std::get<0>(res);
							_state[i].missed(0);
							unsigned int tloc = static_cast<unsigned int>(round(locs(i) * fs / 1000.));
							pm[i][tloc] = ecglib::annotation(tloc,ecglib::annotation_type::RPEAK,i);
							// FIXME: Qon/Qoff

							++nready;
						// Strategy 2: Look to see if any noise peaks are acceptable (>thr2)	
						} else if(num >= 0) {
							if(maxnpeak.value > _state[i].thr2() && maxnpeak.value > _minval) {
								_state[i].reclassify_npeak(num);
				
								auto res = findcog(input.col(i), _frac, _sfrac, maxnpeak.loc);
								// FIXME: Qon/Qoff
								int loc = std::get<0>(res);

								locs(i) = loc;
								++nready;
								_state[i].missed(0);
								// Note: Nothing was found
							}
						}
					}

					if(nready > _nreq) {
						for(std::size_t i = 0; i < _nch; ++i) {
							if(locs(i) > 0) {
								unsigned int tloc = static_cast<unsigned int>(round(locs(i) * 1000./fs));
								pm[i][tloc] = ecglib::annotation(tloc, ecglib::annotation_type::RPEAK, i);
								// FIXME: Qon/Qoff

								if(input(locs(i),i) < _state[i].thr1()) {
									_state[i].spki(_state[i].spki() * _degrad);
									_state[i].npki(_state[i].npki() * _degrad);
								}	
							} else {
								_state[i].missed();
							}
						}

						uvec lidx = find(locs>0);
						vec tmp = detail::grab(locs, lidx);
						std::size_t gloc = median(tmp);

						_lastloc = gloc;
						_firstdet = false;

						unsigned int tloc = static_cast<unsigned int>(round(gloc * 1000./fs));
						pm[ecglib::GLOBAL_LEAD][tloc] = ecglib::annotation(tloc, ecglib::annotation_type::RPEAK, ecglib::GLOBAL_LEAD);
						// FIXME: Qon/Qoff

						_rrpred.detectqrs(gloc, fs);

						for(std::size_t i = 0; i < _nch; ++i) {
							_state[i].ready(false);
						}
					} else {
						// Label as missed so we will not hit this branch again
						_rrpred.missed(idx, fs);

						// Label leads with no detections as missing
						for(std::size_t i = 0; i < _nch; ++i) {
							if(locs(i) == 0) {
								_state[i].missed();
							}
						}
					}
				// Case 3: If we have had no detections in a while lower all thresholds and we are not ready
				} else if (!_ready && ((!_firstdet && (idx - _lastloc)>missedtime_samp) || (_firstdet && idx > missedtime_samp))) {
					unsigned int nready = 0;

					for(std::size_t i = 0; i < _nch; ++i) {
						if(_state[i].isready()) {
							nready++;
						}
					}

					if(nready == 0) {
						for(std::size_t i = 0; i < _nch; ++i) {
							_state[i].missed();
						}
						
						_lastloc = idx;
					}
				}
			}

		public:
			/**
			* @brief Force signal level for channel ch
			*
			* @param ch Channel ch
			* @param val Signal level
			*/
			void spki(const std::size_t ch, const double val) {
				_state[ch].spki(val);
			}

			/**
			* @brief Get signal level for channel ch
			*
			* @param ch Channel ch
			*
			* @return Signal level
			*/
			double spki(const std::size_t ch) const {
				return _state[ch].spki();
			}

			/**
			* @brief Force noise level for channel ch
			*
			* @param ch Channel ch
			* @param val Noise level
			*/
			void npki(const std::size_t ch, const double val) {
				_state[ch].npki(val);
			}

			/**
			* @brief Get noise level for channel ch
			*
			* @param ch Channel ch
			*
			* @return Noise level
			*/
			double npki(const std::size_t ch) const {
				return _state[ch].npki();
			}

		protected:
			/**
			* @brief number of channels
			*/
			std::size_t _nch;

			/**
			* @brief Detector state by channel
			*/
			std::vector<qrsdetector_state> _state;

			/**
			* @brief If we are ready
			*/
			bool _ready;

			/**
			* @brief First detected (by lead) QRS detection in this QRS complex (samples)
			*/
			std::size_t _startloc;

			/**
			* @brief Refactory period of detector (ms)
			*/ 
			std::size_t _minrr;

			/**
			* @brief Max distance we can move when detecting center of gravity (ms)
			*/
			std::size_t _maxmove;

			/**
			* @brief Minimum QRS width (ms)
			*/
			std::size_t _minwidth;

			/**
			* @brief Maximum QRS width (ms)
			*/
			std::size_t _maxwidth;

			/**
			* @brief Minimum left width (ms)
			*/
			std::size_t _minleft;

			/**
			* @brief Minimum right width (ms)
			*/
			std::size_t _minright;

			/**
			* @brief Search window to look for secondary peaks in other leads when a global detection has been made (ms)
			*/
			std::size_t _searchw;

			/**
			* @brief Minimum signal to noise ratio to detect candidate peaks
			*/
			double _minsnr;

			/**
			* @brief Peaks greater than maxdet * signal level are ignored
			*/
			double _maxdet;

			/**
			* @brief U3 value fraction
			*/
			double _frac;

			/**
			* @brief U3 slope fraction
			*/
			double _sfrac;

			/**
			* @brief Degrading signal and noise level (i.e. primary and secondary threshold) by this fraction when secondary detections are made in all leads (using RR prediction)
			*/
			double _degrad;

			/**
			* @brief If this is first detection
			*/
			bool _firstdet;

			/**
			* @brief Last detected peak
			*/
			std::size_t _lastloc;

			/**
			* @brief Number of required channels to trigger global detection
			*/
			std::size_t _nreq;

			/**
			* @brief Channels with more than maxinitmissed missed detections cannot trigger a global detection start
			*/
			int _maxinitmissed;

			/**
			* @brief No detections for missedtime (ms) will result in threshold lowering
			*/
			int _missedtime;

			/**
			* @brief Minimum U3 value to allow
			*/
			double _minval;

			/**
			* @brief Flag if we are storing left/right of candidate peaks as on/offset
			*/
			int _detectonoff;

			/**
			* @brief RR predictor policy
			*/
			rrpredpolicy _rrpred;
	};

	/**
	* @brief Used to store qrs detector debugging values, updated for each candidate peak
	*/
	struct qrsdebug {
		/**
		* @brief Current location of candidate peak
		*/
		std::size_t loc;

		/**
		* @brief Signal level
		*/
		double spki;

		/**
		* @brief Noise level
		*/
		double npki;

		/**
		* @brief Primary threshold
		*/
		double thr1;

		/**
		* @brief Secondary threshold
		*/
		double thr2;

		/**
		* @brief Next QRS according to RR predictor
		*/
		std::size_t nextqrs;

		/**
		* @brief Center of gravity
		*/
		std::size_t gloc;

		/**
		* @brief Left of candidate peak
		*/
		std::size_t left;

		/**
		* @brief Right of candidate peak
		*/
		std::size_t right;
	};

	/**
	* @brief QRS debugger
	*
	* @tparam rrpredpolicy RR prediction policy
	*/
	template<class rrpredpolicy>
	class qrsdetectorpolicy_debug : public qrsdetectorpolicy<rrpredpolicy> {
		public:
			/**
			* @brief Init qrs debugger
			*
			* @param rrpred RR predictor
			* @param nchannels Number of channels
			* @param cfg Configuration
			*/
			qrsdetectorpolicy_debug(const rrpredpolicy rrpred, const std::size_t nchannels, const qrsdetector_config &cfg) : qrsdetectorpolicy<rrpredpolicy>(rrpred, nchannels, cfg) {
				_dbg.resize(nchannels);
			}

		public:
			/**
			* @brief Detect candidate peak
			*
			* @tparam data Data type
			* @param state Detector state
			* @param input Input data
			* @param ch Channel
			* @param idx Index of candidate peak
			* @param fs Sampling frequency
			*
			* @return True if candidate peak was accepted
			*/
			template<class data>
			bool cpeak(detectorstate &state, const data &input, const std::size_t ch, const std::size_t idx, const double fs) {
				qrsdebug dbg;
				dbg.loc = state.cpeak(ch);
				dbg.spki = this->_state[ch].spki();
				dbg.npki = this->_state[ch].npki();
				dbg.thr1 = this->_state[ch].thr1();
				dbg.thr2 = this->_state[ch].thr2();

				if(this->_rrpred.ready()) {
					dbg.nextqrs = this->_rrpred.nextqrs(fs);
				} else {
					dbg.nextqrs = 0;
				}
				bool res = qrsdetectorpolicy<rrpredpolicy>::cpeak(state, input, ch, idx, fs);

				if(res) {
					dbg.gloc = this->_state[ch].location();	
					dbg.left = this->_state[ch].left();
					dbg.right = this->_state[ch].right();
				} else {
					dbg.gloc = 0;
					dbg.left = 0;
					dbg.right = 0;
				}

				_dbg[ch].push_back(dbg);

				return true;
			}

			/**
			* @brief Get debuginning results for channel ch
			*
			* @param ch Channel ch
			*
			* @return Debug matrix
			*/
			arma::mat debug(const std::size_t ch) const {
				arma::mat out = arma::zeros<arma::mat>(_dbg[ch].size(),9);

				for(std::size_t i = 0; i < _dbg[ch].size(); ++i) {
					out(i,0) = _dbg[ch][i].loc;
					out(i,1) = _dbg[ch][i].spki;
					out(i,2) = _dbg[ch][i].npki;
					out(i,3) = _dbg[ch][i].thr1;
					out(i,4) = _dbg[ch][i].thr2;
					out(i,5) = _dbg[ch][i].nextqrs;
					out(i,6) = _dbg[ch][i].gloc;
					out(i,7) = _dbg[ch][i].left;
					out(i,8) = _dbg[ch][i].right;
				}

				return out;
			}

		public:

		protected:
			/**
			* @brief Stored debugging data
			*/
			std::vector<std::vector<qrsdebug>> _dbg;
	};

	/**
	* @brief QRS detector class
	*
	* @tparam detectorpolicy Detector policy
	*/
	template<class detectorpolicy>
	class qrsdetector {
		public:
			/**
			* @brief Init qrs detector
			*
			* @param nchannels Number of channels
			* @param cfg Configuration
			* @param detpol Detector policy
			*/
			qrsdetector(const std::size_t nchannels, const qrsdetector_config &cfg, const detectorpolicy &detpol) : _detectorpol(detpol), _nch(nchannels), _detectorstate(nchannels), _trainpol(nchannels,cfg) {
				_minval = cfg.get<double>("minval");
				_maxval = cfg.get<double>("maxval");
				_trainfrac = cfg.get<double>("trainfrac");

				_minrr = cfg.get<int>("minrr");
				_frac = cfg.get<double>("detfrac");
			}

		public:
			/**
			* @brief Train QRS detector (state after training is discarded)
			*
			* @tparam data Data type
			* @param input Input data
			* @param nsamp Number of samples to use
			* @param nch Number of channels
			* @param fs Sampling frequency
			*/
			template<class data>
			void train(const data &input, const std::size_t nsamp, const std::size_t nch, const double fs) {
				detectorstate train_state(_nch);
				std::size_t minrrsamp = static_cast<std::size_t>(round(_minrr * fs/1000.));

				auto res = detect_peaks(input, nsamp, nch, fs, train_state, _trainpol, _frac, _minrr);
				auto pm = std::get<0>(res);

				const double min = _minval;
				const double max = _maxval;
				const double train_frac = _trainfrac;
	
				for(std::size_t i = 0; i < _nch; ++i) {
					auto aset = pm.find(i);

					if (aset != pm.end()) {
						// Step1: Copy peaks over if they are within min/max
						arma::vec tp = arma::zeros<arma::vec>(aset->second.size());
						std::size_t cur = 0;

						for(auto &ann : aset->second) {
							double val = input(ann.first, i);

							if(val > min && val < max) {
								tp(cur) = val;
								++cur;
							}
						}

						if(cur > 0) {
							// Step2: Determine signal / noise level
							tp = tp.subvec(0,cur-1);

							const double thr = train_frac * median(tp);

							uvec idx = find(tp > thr);
							double sig = 0;
							if (idx.n_elem > 0) {
								vec tmp = detail::grab(tp, idx);
								sig = median(tmp);
							}

							idx = find(tp < thr);
							double noise = 0;
							if (idx.n_elem > 0) {
								vec tmp = detail::grab(tp, idx);
								noise = median(tmp);
							}

							_detectorpol.spki(i, sig);
							_detectorpol.npki(i, noise);
						} else {
							// NOTE: Design choice to allow this to happen but the resulting detector could be overly sensitive
							_detectorpol.spki(i, 0);
							_detectorpol.npki(i, 0);
						}
					}
				}
			}

			/**
			* @brief Get QRS detector policy
			*
			* @return QRS detector policy
			*/
			detectorpolicy policy() const {
				return _detectorpol;
			}

			/**
			* @brief Detect QRS complexes (state is stored afterwards)
			*
			* @tparam data Data type
			* @param input Input data
			* @param nsamp Number of samples
			* @param nch Number of channels
			* @param fs Sampling frequency
			*
			* @return Annotations
			*/
			template<class data>
			typename detectorpolicy::annotation detect(const data &input, const std::size_t nsamp, const std::size_t nch, const double fs) {
				// Step1: Detect peaks using current state
				auto res = detect_peaks(input, nsamp, nch, fs, _detectorstate, _detectorpol,_frac, _minrr);

				// Step2: Update state
				_detectorstate = std::get<1>(res);
				_detectorpol = std::get<2>(res);

				// Step3: Return detections
				return std::get<0>(res);
			}

		protected:
			/**
			* @brief Detector policy
			*/
			detectorpolicy _detectorpol;

			/**
			* @brief Detector state
			*/
			detectorstate _detectorstate;

			/**
			* @brief Training policy
			*/
			qrstrainingpolicy _trainpol;

			/**
			* @brief Number of channels
			*/
			std::size_t _nch;

			/**
			* @brief Minimum value tolerated
			*/
			double _minval;

			/**
			* @brief Maximum value tolerated
			*/
			double _maxval;

			/**
			* @brief Fraction of median peaks in training to use for initial signal level
			*/
			double _trainfrac;

			/**
			* @brief Used to trigger candidate peak detections
			*/
			double _frac;

			/**
			* @brief Refactory period
			*/
			std::size_t _minrr;
	};
}

#endif
