/**
 * @file qrsdetector.hpp
 * @author Lars Johannesen <lj@halloffame.dk>
 *
 * @section LICENSE
 * This program is licensed under LGPL3, see LICENSE
 * 
 * @section DESCRIPTION
 * The classes in this file are used for describing characteristics about ecg records
 */
#ifndef ECGLIB_QRSDETECTOR_U3_LJ_2014_06_06
#define ECGLIB_QRSDETECTOR_U3_LJ_2014_06_06 1

#include <ecglib_bits/qrsdetectors/qrsdetector.hpp>
#include <ecglib_bits/qrsdetectors/detail/detail.hpp>
#include <ecglib_bits/util/config.hpp>

#include <boost/circular_buffer.hpp>

namespace ecglib {
	/**
	* @brief Calculates U3 sample-by-sample
	*
	* @tparam ITERIN Iterator type
	* @tparam ITEROUT Iterator type
	* @param start Start iterator
	* @param stop Stop iterator
	* @param out Output iterator
	* @param U3N Width (ms)
	*/
	template <class ITERIN, class ITEROUT>
	void calcU3(ITERIN start, ITERIN stop, ITEROUT out, int U3N=100, int ith=2) {
		if ( (U3N % 2) != 0) {
			throw std::logic_error("Bad_Window_Size");
		}

		int nsamples = stop-start;

		int zeropad = U3N+1;
		vec data = zeros<vec>(zeropad*2+nsamples);
		vec outdata = zeros<vec>(zeropad*2+nsamples);
		ITERIN iter=start;
		int i = 0;
		while(iter != stop) {
			data[zeropad+i] = *iter;

			++iter;
			++i;
		}

		double sum = 0;
		for(int i = ith; i < U3N/2; ++i) {
			sum += pow(data[i]-data[i-ith],2);
		}
		outdata[0] = sum;

		for(int i = 1; i < (nsamples+zeropad*2); ++i) {
			double subterm = 0;
			if(i > U3N/2) {
				subterm = pow(data[i-U3N/2]-data[i-U3N/2+ith],2);
			}

			double addterm = 0;
			if(i < (nsamples+zeropad*2-U3N/2)) {
				addterm = pow(data[i+U3N/2]-data[i+U3N/2-ith],2);
			}

			outdata[i] = outdata[i-1] + addterm - subterm;
		}

		i = 0;
		while(i<nsamples) {
			out[i] = outdata[zeropad+i];

			++i;
		}
	}

	/**
	* @brief Calculate U3 for data
	*
	* @tparam data Data type
	* @param e Data input
	* @param U3N With of U3 window in samples
	* @param ith Ith factor
	*
	* @return Matrix (U3 by lead)
	*/
	template<class data>
	arma::mat calcU3(const data &e, const int U3N=100, const int ith=2) {
		int nsamples = e.n_rows;
		int nleads = e.n_cols;

		arma::mat out(nsamples,nleads);

		for(int i = 0; i < nleads; ++i) {
			calcU3(e.begin_col(i),e.end_col(i),out.begin_col(i),U3N,ith);
		}

		return out;
	}

	/**
	* @brief Calculate U3 for data
	*
	* @tparam data Data type
	* @param e Input data
	* @param fs Sampling frequency 
	* @param cfg Configuration
	*
	* @return Matrix (U3 by lead)
	*/
	template<class data>
	arma::mat calcU3(const data &e, const double fs, const qrsdetector_config &cfg) {
		const int u3 = static_cast<int>(cfg.get<int>("u3window")*fs/1000.);
		const int ith = cfg.get<int>("ith");

		return calcU3(e, u3, ith);
	}
}

#endif
