%% @file 
% eeg_detector.m
%
% @author Lars Johannesen
%
% @section DISCLAIMER
% This software and documentation (the "Software") were developed at the Food
% and Drug Administration (FDA) by employees of the Federal Government in the
% course of their official duties. Pursuant to Title 17, Section 105 of the
% United States Code, this work is not subject to copyright protection and is
% in the public domain. Permission is hereby granted, free of charge, to any
% person obtaining a copy of the Software, to deal in the Software without
% restriction, including without limitation the rights to use, copy, modify,
% merge, publish, distribute, sublicense, or sell copies of the Software or
% derivatives, and to permit persons to whom the Software is furnished to do
% so. FDA assumes no responsibility whatsoever for use by other parties of
% the Software, its source code, documentation or compiled executables, and
% makes no guarantees, expressed or implied, about its quality, reliability,
%  or any other characteristic. Further, use of this code in no way implies
% endorsement by the FDA or confers any advantage in regulatory decisions.
% Although this software can be redistributed and/or modified freely, we
% ask that any derivative works bear some notice that they are derived from
% it, and any modified versions bear some notice that they have been modified.
%
% @section Examples:
%  [idx] = eeg_detector(eeg, anns);
%
% @section See also:
% qrsdetect
%

%% 
% @brief Detects the beats in the EEG/EOG signal using trusted annotations to train the detector
% @param eeg EEG/EOG signal in 250 Hz
% @param anns Trusted detections
% @return onset_time Beat annotations in ms
%
function [idx] = eeg_detector(eeg, anns)
%eeg_detector Detects the beats in the EEG/EOG signal using trusted locations
%
% eeg_detector(abp, anns)
%
% eeg - vector (250 Hz)
% anns - annotations
%
%                                  DISCLAIMER
% This software and documentation (the "Software") were developed at the Food
% and Drug Administration (FDA) by employees of the Federal Government in the
% course of their official duties. Pursuant to Title 17, Section 105 of the
% United States Code, this work is not subject to copyright protection and is
% in the public domain. Permission is hereby granted, free of charge, to any
% person obtaining a copy of the Software, to deal in the Software without
% restriction, including without limitation the rights to use, copy, modify,
% merge, publish, distribute, sublicense, or sell copies of the Software or
% derivatives, and to permit persons to whom the Software is furnished to do
% so. FDA assumes no responsibility whatsoever for use by other parties of
% the Software, its source code, documentation or compiled executables, and
% makes no guarantees, expressed or implied, about its quality, reliability,
%  or any other characteristic. Further, use of this code in no way implies
% endorsement by the FDA or confers any advantage in regulatory decisions.
% Although this software can be redistributed and/or modified freely, we
% ask that any derivative works bear some notice that they are derived from
% it, and any modified versions bear some notice that they have been modified.
%
% Written by: Lars Johannesen
%
% Examples:
%  [idx] = eeg_detector(abp, anns);
%
% See also:
% qrsdetect

	% Make sure anns is a rowvector
	anns = anns(:);
	anns = anns';

	% Set Fs fixed for now
	Fs = 250; % Hz
	idx = []; % Empty indices 

	% Threshold for an EEG/EOG annotation to match an annotation in the annotation set
	thr = round(50 * Fs/1000.); % 80 msec

	% Requirements for detected peaks
	minwidth = round(20 * Fs/1000.); % 20 msec
	maxwidth = round(200 * Fs/1000.); % 200 msec

	% We need the annotations in samples
	anns = round(anns * Fs/1000.); % Convert to samples

	% Preprocess the EEG/EOG signal
	[b,a] = butter(5, [5 35]/(Fs/2.));
	eeg2 = filtfilt(b,a,eeg);	
	u3 = calcu3(eeg2, Fs, 64, 8);

	% To train the detector we:
	%  - Locate regions where we have reliable annotations (gapless regions)
	%  - For all gapless regions:
	%    Determine which EEG/EOG peaks match annotations (signal peaks)
	%    The remaining peaks are (noise peaks)
	%    Detect based on signal and noise peaks in the whole record\

	% Determine regions of trustworthy annotations of durations > 60s

	% Gap definition
	mingap = 1500; % msec

	% Minimum duration for a region to be used for training
	minreg = 60; % sec

	% Compute RR series
	rr = [diff(anns) 0]; % To maintain relationship to rr

	% Locate gaps
	gaps = find(rr > mingap);

	if isempty(gaps)
		% Setup matrix of regions
		regions = [anns(1) anns(end) anns(end)-anns(1)];
	else
		% Setup matrix of regions
		regions = zeros(length(gaps)+1,3);

		% First region starts at annotation 1, the remaining starts before the gap
		regions(:,1) = [anns(1) anns(gaps-1)];

		% The first n-1 regions stop right after the gap, but the last region stops at the
		% last annotation
		regions(:,2) = [anns(gaps) anns(end)];

		% Compute region lengths in seconds
		regions(:,3) = (regions(:,2) - regions(:,1)) / 1000.;
	end

	% Grab regions that have satisfy min length
	train_regions = regions(regions(:,3)>minreg,:);
	[nreg,~] = size(train_regions);

	if nreg == 0
		idx = [];
	else
		% Detect peaks
		all_peaks_in = detect_peaks(u3, Fs);

		[npeaks,~] = size(all_peaks_in);
		if npeaks > 0 
			% Compute widths
			widths = all_peaks_in(:,3) - all_peaks_in(:,2);
			
			% Good peaks satisfies all
			goodpeaks = widths>minwidth&widths<maxwidth;

			% All peaks we can use are the good peaks and grab the center of gravitiy location
			all_peaks = all_peaks_in(goodpeaks,1);

			% Grab all peaks that occur within a region we use for training
			train_peaks = zeros(length(all_peaks),2);
			train_peaks(:,1) = all_peaks;
			[nreg,~] = size(train_regions);

			for i = 1:nreg
				idx = train_peaks(:,1)>train_regions(i,1)&train_peaks(:,1)<train_regions(i,2);
				train_peaks(idx,2) = 1;
			end

			train_peaks = train_peaks(train_peaks(:,2)==1,1);

			% Determine which of the training peaks are close to trusted annotations
			peaks_classes = zeros(length(train_peaks),1);

			% NOTE: Peak times is more for debugging purposes
			peaks_times = zeros(length(train_peaks),1);
			for i=1:length(peaks_classes)
				[val,cidx] = min(abs(train_peaks(i)-anns));

				if val < thr
					peaks_classes(i) = 1;
					peaks_times(i) = val;
				end
			end

			% How many peaks occured close to a trusted annotation?
			npeaks = sum(peaks_classes==1);

			% Based on the training set it seems reasonable to smooth the detection threshold
			% over 50 detections
			nfilt = 50;

			% If we for some reason had a short record, lets allow a smaller window
			if npeaks < nfilt
				nfilt = npeaks-1;
			end

			% Less than 2 peaks ? Skip as we cant interp it and its meaningless
			if npeaks > 1
				% Compute signal threshold using linear interpolation
				sigthr = interp1(train_peaks(peaks_classes==1),u3(train_peaks(peaks_classes==1)), all_peaks);

				% If for some reason we have too few noise peaks to interp the noisethr
				% lets just use a repeated average value
				noisenpeaks = sum(peaks_classes==0);
				if noisenpeaks > 1
					noisethr = interp1(train_peaks(peaks_classes==0),u3(train_peaks(peaks_classes==0)), all_peaks);
				else
					[r,c] = size(all_peaks);
					noisethr = rep(mean(u3(train_peaks(peaks_classes==0))),r,c);
				end

				% Smooth the interpolated thresholds
				sigthr2 = filter(ones(nfilt,1)/nfilt,1, sigthr);
				noisethr2 = filter(ones(nfilt,1)/nfilt,1, noisethr);

				%plot(all_peaks, sigthr2,'g');hold on;plot(u3, 'b');plot(anns, u3(anns),'g*');plot(all_peaks,u3(all_peaks)*1.1,'r*');plot(all_peaks, noisethr2,'r');hold off;
				
				% Determine a detection threshold that floats above noisethr
				detthr = (sigthr2 - noisethr2) * 0.5 + noisethr2;

				% Compute a crude metric of SNR
				sqi = sigthr2 ./ noisethr2;

				% Grab those peaks (from the entire record) that are above detection
				% threshold and have a snr > 1.5
				idx = all_peaks(u3(all_peaks)>detthr&sqi>1.5);

				% Compute how many detections we had
				coverage = length(idx) / length(anns);

				% If the EEG/EOG detects fewer than 20% of the overall detections
				% lets just ignore it.. its likely noise
				if coverage < 0.5
					idx = [];
				end
			else
				idx = [];
			end
		end
	end
end
