%% @file
% majorityvote.m
%
% @author Loriano Galeotti
%
% @section DISCLAIMER
% This software and documentation (the "Software") were developed at the Food
% and Drug Administration (FDA) by employees of the Federal Government in the
% course of their official duties. Pursuant to Title 17, Section 105 of the
% United States Code, this work is not subject to copyright protection and is
% in the public domain. Permission is hereby granted, free of charge, to any
% person obtaining a copy of the Software, to deal in the Software without
% restriction, including without limitation the rights to use, copy, modify,
% merge, publish, distribute, sublicense, or sell copies of the Software or
% derivatives, and to permit persons to whom the Software is furnished to do
% so. FDA assumes no responsibility whatsoever for use by other parties of
% the Software, its source code, documentation or compiled executables, and
% makes no guarantees, expressed or implied, about its quality, reliability,
%  or any other characteristic. Further, use of this code in no way implies
% endorsement by the FDA or confers any advantage in regulatory decisions.
% Although this software can be redistributed and/or modified freely, we
% ask that any derivative works bear some notice that they are derived from
% it, and any modified versions bear some notice that they have been modified.
%

%%
% @brief Generate moajority vote amongs multiple annotations
% @param anns - Annotations (in ms) (cell array of column vector)
% @param options struct
%       options.absthresh = absolute threshold (e.g. 2 means that >= 2 annotatopms happen at the same time  the majority votes for annotation present) The final threshold is the minimum between the absolute and relative!
%       options.relthresh = relative threshold (e.g. 0.5 means that >= half annotations happen at the same time  the majority votes for annotation present)  The final threshold is the minimum between the absolute and relative!
%       options.emptychannelthresh = minimum number of annotations needed in a channel to consider it "active" (i.e. counted as present for computing the relative threshold). The annotations in inactive channels are however used.
%       options.wintype = type of window for the timeserie. Can be 'gauss' or 'gaussian' for gaussian window, otherwise the default is tukey.
%       options.winsize = first parameter for gauss and tukey window (basically is the size of the window.
%       options.windecay = second parameter for gauss and tukey window (basically is the decay lenght).
%       options.winwave = []  or {ii} optional window waveform, by signal(if is a cell array of same dimensions as anns) or common for all signal (if is a vector). If empty, or the size is different than the number of signals it is overridden using wintype/winsize/windecay, otherwise is used instead of global window definition.
%       options.sf  =  sampling freq for the annotations, default 1000 .
%       options.len  =  lenght of the output signal .
%       options.verbose = verbosity level. 0 no output; 1 = warning when rollbacking to safety defaults
%       options.debugplot = set at 1 to show debug plots
%       options.mergemargin = margin in ms between multiple output annotations. If multiple annotations are detected within the margin  
%       options.mergestrategy = strategy to merge annotations: 1 = (default) average ; other strategies TBD :) ;     
% @param weight weight of each annotation (single row vector), corresponds to each annotation vector in the cell array. If empty or not matching the size each channel has weight 1 (default)
% @param qual for each annotation the relative quality (reliability), 0 = unreliable, 1 = 100% reliable
% @return [out, options, sig, candidate], out: annotations (in ms), options struct used, sig is majority signal, candidate are candidates in majority signal
%
function [ out , options, sig, candidate ]  = majorityvote(anns,options,weight,qual)
%majorityvote Generate moajority vote amongs multiple annotations
%
% anns - Annotations (in ms) (cell array of column vector)
% options struct
%       options.absthresh = absolute threshold (e.g. 2 means that >= 2 annotatopms happen at the same time  the majority votes for annotation present) The final threshold is the minimum between the absolute and relative!
%       options.relthresh = relative threshold (e.g. 0.5 means that >= half annotations happen at the same time  the majority votes for annotation present)  The final threshold is the minimum between the absolute and relative!
%       options.emptychannelthresh = minimum number of annotations needed in a channel to consider it "active" (i.e. counted as present for computing the relative threshold). The annotations in inactive channels are however used.
%       options.wintype = type of window for the timeserie. Can be 'gauss' or 'gaussian' for gaussian window, otherwise the default is tukey.
%       options.winsize = first parameter for gauss and tukey window (basically is the size of the window.
%       options.windecay = second parameter for gauss and tukey window (basically is the decay lenght).
%       options.winwave = []  or {ii} optional window waveform, by signal(if is a cell array of same dimensions as anns) or common for all signal (if is a vector). If empty, or the size is different than the number of signals it is overridden using wintype/winsize/windecay, otherwise is used instead of global window definition.
%       options.sf  =  sampling freq for the annotations, default 1000 .
%       options.len  =  lenght of the output signal .
%       options.verbose = verbosity level. 0 no output; 1 = warning when rollbacking to safety defaults
%       options.debugplot = set at 1 to show debug plots
%       options.mergemargin = margin in ms between multiple output annotations. If multiple annotations are detected within the margin  
%       options.mergestrategy = strategy to merge annotations: 1 = (default) average ; other strategies TBD :) ;     
% weight weight of each annotation (single row vector), corresponds to each annotation vector in the cell array. If empty or not matching the size each channel has weight 1 (default)
% qual for each annotation the relative quality (reliability), 0 = unreliable, 1 = 100% reliable
%
%                                  DISCLAIMER                                  
% This software and documentation (the "Software") were developed at the Food  
% and Drug Administration (FDA) by employees of the Federal Government in the  

% United States Code, this work is not subject to copyright protection and is  
% in the public domain. Permission is hereby granted, free of charge, to any   
% person obtaining a copy of the Software, to deal in the Software without     
% restriction, including without limitation the rights to use, copy, modify,   
% merge, publish, distribute, sublicense, or sell copies of the Software or    
% derivatives, and to permit persons to whom the Software is furnished to do   
% so. FDA assumes no responsibility whatsoever for use by other parties of     
% the Software, its source code, documentation or compiled executables, and    
% makes no guarantees, expressed or implied, about its quality, reliability,   
%  or any other characteristic. Further, use of this code in no way implies    
% endorsement by the FDA or confers any advantage in regulatory decisions.     
% Although this software can be redistributed and/or modified freely, we       
% ask that any derivative works bear some notice that they are derived from    
% it, and any modified versions bear some notice that they have been modified. 
%                                                                              
% Written by: Loriano Galeotti

% input check and set defaults
totchannels = length(anns);

if nargin <2 || ~isstruct(options)
    options = struct('absthresh',[],'relthresh',[], 'emptychannelthresh', []  );
end
if nargin <3
    weight = [];
end
if nargin <4 || ~iscell(qual)
    qual = {};
    qual(1:totchannels)={[]};
end

% robustly detect last annotation across channels
maxann = ((cellfun(@max, anns,'uniformoutput', false )));
lastann = max(cell2mat(maxann(not(cellfun('isempty', maxann)))));

% setting defaults
options = fieldscheckeasy(options, 'verbose', 0 , 'absthresh',2,'relthresh',.5, 'emptychannelthresh', 10, 'winsize', 100, 'windecay', .75 , 'wintype', 'tukey','sf',1000, 'len', lastann + 500 ,'winwave', [], 'verbose',1, 'debugplot',0,'mergemargin',100 ,'mergestrategy',1,'mergeampdiff',0.1);
if length(weight)~=totchannels
    weight=ones(size(anns));
end
if length(qual)~=totchannels
    qual{1:totchannels}=[]; % wrong size for quality, force it empty
end

% check if window is provided or need to be created internally
if iscell(options.winwave) % check if winwave is valid
    if length(options.winwave)~=totchannels% is a cell array, check if the size is correct otherwise roll back to the default window
        if options.verbose
            warning('Majorityvote: options.winwave is a cell, size does not match with anns. Using global window instead by-lead window')
        end
        options.winwave=[];
    end
end

if isempty(options.winwave) % generate waves if needed
    % not valid window provided, generate one
    % correct window paramenters
    options.winsize= max(options.winsize,2) ;
    if mod(options.winsize,2)~=0
        options.winsize= options.winsize+1; % force winsize to be even (otherwise make_series complains)
    end
    options.windecay= min(options.windecay,1) ;
    options.windecay= max(options.windecay,0) ;
    options.winwave = {};
    switch lower(options.wintype)
        case {'gauss', 'gaussian'}
            options.winwave(1:totchannels) = {gaussian( options.winsize, options.windecay)}; % Width in ms - create gaussian mixing window
        otherwise % use tukey
            options.winwave(1:totchannels) = {tukeywin( options.winsize, options.windecay)};  % Width in ms - create tukey mixing window.  first param is the total window length in ms. The second param is the cumulative ration the tapered cosine part (2 sided decay), so the function stays at 1 for 1-param2.
    end
    
elseif isnumeric(options.winwave)
    % numeric winwave, apply the same to all signals
    temp = options.winwave;
    options.winwave= {};
    options.winwave(1:totchannels)={temp};
end

% detect active channels and set detection threshold
activechannels = sum(cellfun(@numel, anns)>options.emptychannelthresh);  % collect channels with enough detections.
options.threshold = min(options.absthresh, options.relthresh*activechannels) ; % compute threshold

if options.debugplot ; clf; end  % prepare the plot in case of need

%% create cumulative waveform
sig =[];
for jj = 1:totchannels
    if isempty(sig)
        sig = make_series(anns{jj},options.winwave{jj},options.len,qual{jj})*weight(jj) ; % cumulative signal, wighted by channel and with signal quality applied
    else
        sig = sig + make_series(anns{jj},options.winwave{jj},options.len,qual{jj})*weight(jj) ; % cumulative signal, wighted by channel and with signal quality applied
    end
    if options.debugplot
        hold all ;
        plot (make_series(anns{jj},options.winwave{jj},options.len,qual{jj})*weight(jj)+jj )
    end
end
%% search detections
% search for upcrossing and downcrossing of the threshold
if (sig(1))>=options.threshold % if the signal starts above the thteshold add first detection at 1
    candidate(:,1)= [1; find((sig(1:end-1)<options.threshold )&(sig(2:end)>=options.threshold ))];
else
    candidate(:,1)= find((sig(1:end-1)<options.threshold )&(sig(2:end)>=options.threshold ));
end
if (sig(end))>=options.threshold % if the signal ends above the thrteshold add last detec iotn at signal end (should not happen as the detection signal is longer than the last detection )
    candidate(:,3)=  [find((sig(2:end)<options.threshold )&(sig(1:end-1)>=options.threshold ) )+1 ; length(sig)];
else
    candidate(:,3)=  [find((sig(2:end)<options.threshold )&(sig(1:end-1)>=options.threshold ) )+1 ];
end
candidate(:,2) = ceil(candidate(:,1)+(candidate(:,3)-candidate(:,1))/2); % get centers of the detections

candidate(:,4) = sig(candidate(:,2));

%% merge detections if needed
if options.mergemargin % check if is needed to merge detections
    jj = size(candidate,1); % loop backward the detections
    while jj > 1
        if abs(candidate(jj,2)-candidate(jj-1,2))<=options.mergemargin % check if a detection is too close to the previous

		ratio = candidate(jj,4) / candidate(jj-1,4);

		if abs(1-ratio) < options.mergeampdiff
		    candidate(jj-1,2)= round(candidate(jj-1,2)+(candidate(jj,2)-candidate(jj-1,2))/2); % get centers of the detections
		    candidate(jj-1,3)= candidate(jj,3); %assign the end as it was the end of the detection
		else
			if candidate(jj-1,4) < candidate(jj,4)
				candidate(jj-1,:) = candidate(jj,:);
			end
		end
		candidate(jj-1,4) = sig(candidate(jj-1,2));
	        candidate(jj,:) = []; % remove removed beat
        end
        jj = jj-1;
    end
end
%% output
% prepare outputs
out = candidate(:,2);

%% debug plot if needed
if options.debugplot
    plot(sig,'k.') ;
    plot(candidate(:,1),ones(size(candidate,1),1)*options.threshold,'b^')
    plot(candidate(:,3),ones(size(candidate,1),1)*options.threshold,'bv')
    plot(candidate(:,2),ones(size(candidate,1),1)*options.threshold,'bs')
    xlabel('Voting @ 1 Khz');
    ylabel('channel# ; final score ; up/center/down detections ')
end
end
