%% @file
% qrsdetect.m
%
% @author Lars Johannesen
%
% @section DISCLAIMER
% This software and documentation (the "Software") were developed at the Food  
% and Drug Administration (FDA) by employees of the Federal Government in the  
% course of their official duties. Pursuant to Title 17, Section 105 of the    
% United States Code, this work is not subject to copyright protection and is  
% in the public domain. Permission is hereby granted, free of charge, to any   
% person obtaining a copy of the Software, to deal in the Software without     
% restriction, including without limitation the rights to use, copy, modify,   
% merge, publish, distribute, sublicense, or sell copies of the Software or    
% derivatives, and to permit persons to whom the Software is furnished to do   
% so. FDA assumes no responsibility whatsoever for use by other parties of     
% the Software, its source code, documentation or compiled executables, and    
% makes no guarantees, expressed or implied, about its quality, reliability,   
%  or any other characteristic. Further, use of this code in no way implies    
% endorsement by the FDA or confers any advantage in regulatory decisions.     
% Although this software can be redistributed and/or modified freely, we       
% ask that any derivative works bear some notice that they are derived from    
% it, and any modified versions bear some notice that they have been modified. 
%
% @section Examples:
% cfg = qrsdetect(); % Obtain default config
% [anns,cfg] = qrsdetect(ecg, fs, u3, cfg); % Detect using cfg
% anns = qrsdetect(ecg, fs, u3); % Default cfg detections
%
% @section See also:
% calcu3

%%
% @brief Detects QRS complexes
% @param ecg ECG matrix
% @param fs Sampling frequency
% @param u3 U3 matrix (see calcu3)
% @param cfg Configuration struct (optional)
% @return pm = annotations, cfg as an optional output

function [pm] = qrsdetect(ecg, fs, u3, cfg)
%qrsdetect Detects QRS complexes by lead
%
% qrsdetect(ecg, u3, cfg)
%  ecg           - ECG data matrix
%  fs            - Sampling frequency
%  u3            - U3 matrix (see calcu3)
%  cfg           - optional config struct see below.
%
% The detector works in two steps: training and detection. First, it detects peaks during the training 
% part or segment (training_time). The idea of the training part is to determine the detection thresholds
% and after the training the detector will restart from sample 0 with the training thresholds. The initial
% thresholds are determined by detecting N largest (but below < maxthr) peaks in training, where N is 
% determined by training_time and training_rr.
%
% After training, as described above the detector restarts with the thresholds determined by training
% and detects peaks. The detector keeps per lead a history of detected signal (>threshold) and noise (<threshold)
% peaks.
%
% Detects QRS complexes using the u3 signal. The cfg argument is a struct with the following fields:
%  minthr        - Minimum peak to be detected in training part
%  maxthr        - Maximum peak to be detected in training part
%  minrr         - Minimum RR interval
%  training_time - Training time (ms)
%  training_rr   - Training RR (ms)
%  history       - History of peaks
%  nbeats        - Used for median beats (nbeats=1 otherwise nbeats=-1)
%  reqpeaks      - Number of peaks required to trigger a detection 
%  mintwave      - If a peak is detected within mintwave ms of previous and is <50% of the size
%                  it is probably a tall T-wave
%  frac1         - primary threshold is noiselevel + frac1 * (signallevel-noiselevel)
%                  where signallevel and noiselevel are estimated by the detector
%  frac2         - secondary threshold is primary threshold * 1/2
%                  (used when missed detections occur)
%
%                                  DISCLAIMER                                  
% This software and documentation (the "Software") were developed at the Food  
% and Drug Administration (FDA) by employees of the Federal Government in the  
% course of their official duties. Pursuant to Title 17, Section 105 of the    
% United States Code, this work is not subject to copyright protection and is  
% in the public domain. Permission is hereby granted, free of charge, to any   
% person obtaining a copy of the Software, to deal in the Software without     
% restriction, including without limitation the rights to use, copy, modify,   
% merge, publish, distribute, sublicense, or sell copies of the Software or    
% derivatives, and to permit persons to whom the Software is furnished to do   
% so. FDA assumes no responsibility whatsoever for use by other parties of     
% the Software, its source code, documentation or compiled executables, and    
% makes no guarantees, expressed or implied, about its quality, reliability,   
%  or any other characteristic. Further, use of this code in no way implies    
% endorsement by the FDA or confers any advantage in regulatory decisions.     
% Although this software can be redistributed and/or modified freely, we       
% ask that any derivative works bear some notice that they are derived from    
% it, and any modified versions bear some notice that they have been modified. 
%                                                                              
% Written by: Lars Johannesen
%                                                                              
% Examples:
%  qrswidth = 100;
%  u3 = calcu3(ecg, qrswidth, 2);
%  cfg = qrsdetect();
%  pm = qrsdetect(ecg, u3, cfg);
%
% See also:
% calcu3
