%% @file
% rrestimate.m
%
% @author Loriano Galeotti
%
% @section DISCLAIMER
% This software and documentation (the "Software") were developed at the Food
% and Drug Administration (FDA) by employees of the Federal Government in the
% course of their official duties. Pursuant to Title 17, Section 105 of the
% United States Code, this work is not subject to copyright protection and is
% in the public domain. Permission is hereby granted, free of charge, to any
% person obtaining a copy of the Software, to deal in the Software without
% restriction, including without limitation the rights to use, copy, modify,
% merge, publish, distribute, sublicense, or sell copies of the Software or
% derivatives, and to permit persons to whom the Software is furnished to do
% so. FDA assumes no responsibility whatsoever for use by other parties of
% the Software, its source code, documentation or compiled executables, and
% makes no guarantees, expressed or implied, about its quality, reliability,
%  or any other characteristic. Further, use of this code in no way implies
% endorsement by the FDA or confers any advantage in regulatory decisions.
% Although this software can be redistributed and/or modified freely, we
% ask that any derivative works bear some notice that they are derived from
% it, and any modified versions bear some notice that they have been modified.
%

%%
% @brief estimates RR serie when there is a gap (likely due a misdetection)
% @param in Input annotations
% @param threshold gap beteen beats
% @param reclen Length of record
% @return out - annotations with gaps filled (using linear interpolation)
%
function [full, fill] = rrestimate (in,  threshold, reclen)
%combined Returns an array containing the locations of the beats detected using the different signals on the record
%
% Input annotations
% threshold between beats
% reclen Length of record
%
%                                  DISCLAIMER
% This software and documentation (the "Software") were developed at the Food
% and Drug Administration (FDA) by employees of the Federal Government in the

% United States Code, this work is not subject to copyright protection and is
% in the public domain. Permission is hereby granted, free of charge, to any
% person obtaining a copy of the Software, to deal in the Software without
% restriction, including without limitation the rights to use, copy, modify,
% merge, publish, distribute, sublicense, or sell copies of the Software or
% derivatives, and to permit persons to whom the Software is furnished to do
% so. FDA assumes no responsibility whatsoever for use by other parties of
% the Software, its source code, documentation or compiled executables, and
% makes no guarantees, expressed or implied, about its quality, reliability,
%  or any other characteristic. Further, use of this code in no way implies
% endorsement by the FDA or confers any advantage in regulatory decisions.
% Although this software can be redistributed and/or modified freely, we
% ask that any derivative works bear some notice that they are derived from
% it, and any modified versions bear some notice that they have been modified.
%
% Written by: Loriano Galeotti

%% input checking and setting defaults 
if nargin <2
    threshold = [];
end

if nargin <3
    reclen = [];
end

if isempty(threshold)
    threshold= 500;
end

%% setting internal parameters 
maxmedianrrdeviation = 1.2 ; % maximum deviation from median rr accepted as rr for computing the filling (helps in case a beat is missed every two)
%maxgapthreshold = 1500 ; % maximum gap allowed between beats
debugplot = 0 ;

%% reshaping inputs and computing derivative and overall statistics 
in = in(:); % ensure input annotation is a column vector
full = in ;
len = length(full);
der = diff(full);
medrr = median(der);
%% old second-derivative search strategy, complex and not satisfacotry
%  secder = diff(der);
% 	gapl = find( abs(secder(1:end-1))<threshold & abs(secder(2:end))>=threshold)  +2;
% 	% gapr = find( (secder(1:end-1)>=threshold & secder(2:end)<threshold) | (secder(1:end-1)<=-threshold & secder(2:end)>-threshold) ) +1;
% 	gapr = find( abs(secder(1:end-1))>=threshold & abs(secder(2:end))<threshold) +1;

%% do gap search and filling

%%% search and detect gaps
gapl = find( abs(der(1:end-1))>threshold )  ;
gapr = gapl+1;

%% fill gap in the begin of recording using the median of the first beats
timegap = in(1);
brl = (median(der(1:3)));
if brl> medrr*maxmedianrrdeviation % if estimated beat rate is eccessive (too low)use median instead
    brl = medrr;
end
fillann = timegap - brl: -brl : 1 ;
fillann = fillann(end:-1:1);
fill = fillann(:);
full = fillann(:);
fillann=[];  %#ok<NASGU>

%% Fill gaps within the annotations
if  isempty(gapl)
    full=[full; in]; % no gap detected, just copy the annotation timeserie
else % if any gap detectd do fill gaps
    if gapr(1)<= gapl(1)
        gapl=[ max(1, gapr(1)-1) ; gapl(:)];  %make sure gap left begins before gap right
    end
    brr = der(end) ;
    for ii = 1 : length(gapl)
        if gapr(ii) < len-3
            brr = min(median(der(gapr(ii):gapr(ii)+2))); % detrmine median beat rate oon the right of the gap
        end
        if gapl(ii) <3
            brl = brr ; % determine beat rate on the left of the gap
        else
            brl = median(der(gapl(ii)-2:gapl(ii))); % determine beat rate on the left of the gap
        end
        
        if brl> medrr*maxmedianrrdeviation % if estimated beat rate is eccessive (too low)use median instead
            brl = medrr;
        end
        if brr> medrr*maxmedianrrdeviation % if estimated beat rate is eccessive (too low)use median instead
            brr = medrr;
        end
        timegap = in(gapr(ii))-in(gapl(ii)) ;
        
        brm = (brl +brr ) /2 ; % get a average beat rate
        % check how many beats to fill
        beatstofill = (timegap/brm)-1;
        
        if  round(beatstofill)> 3 % if more than 3 beats to fill fill linearly, however do not place a beat more than 1/2 beat toward the end
            fillann = round(timegap/round(beatstofill)):round(timegap/round(beatstofill)):timegap-round(timegap/round(beatstofill)/2);
        else
            fillann = [brm:brm:timegap-ceil(brm/2)];
        end
        if ii >1
            full = [full; in(gapr(ii-1):gapl(ii)); fillann(:)+in(gapl(ii)) ] ; %#ok<AGROW>
        else
            full = [full; in(1:gapl(ii)); fillann(:)+in(gapl(ii)) ] ; %#ok<AGROW>
        end
        fill = [fill; fillann(:)];  %#ok<AGROW>
    end
    if isempty(gapr)
        full = [full; in(:) ] ;
    else
        full = [full; in(gapr(ii):end) ] ;
    end
end

%% fill gap in the end of recording using the median of the last beats (if record lenght was specified)
if ~isempty(reclen)
    timegap = reclen - full(end);
    brl = median(der(end-3:end));
    fillann = brl : brl : timegap -1 ;
    full = [full ; full(end)+fillann(:)];
    fill = [fill; fillann(:)];
end

%% Plots for debugging (usually disabled)
if debugplot
    % debug plot
    clf; hold on ; %#ok<UNRCH>
    plot(in,zeros(length(in)),'r*')
    plot(in(gapl), zeros(length(gapl))+.01,'g^')
    plot(in(gapr), zeros(length(gapr))-.01,'rv')
    plot(in(1:length(der)),der/max(der),'c');
    plot(in(1:length(secder)),secder/max(secder),'m');
    plot(fill,zeros(length(fill))+.01,'b.')
    plot(full,zeros(length(full)),'b.')
    plot(full(1:end-1),diff(full)/max(der),':g')
end
end
